import {
    DifficultyCalculator,
    DifficultyHitObject,
} from "@rian8337/osu-difficulty-calculator";
import {
    DifficultyCalculator as RebalanceDifficultyCalculator,
    DifficultyHitObject as RebalanceDifficultyHitObject,
} from "@rian8337/osu-rebalance-difficulty-calculator";
import { BeatmapSection } from "./data/BeatmapSection";

/**
 * A beatmap section generator that generates beatmap section based on aim/speed strain.
 */
export abstract class BeatmapSectionGenerator {
    /**
     * Generates `BeatmapSection`s for the specified beatmap.
     *
     * @param calculator The difficulty calculator to generate for.
     * @param minSectionObjectCount The maximum delta time allowed between two beatmap sections.
     * Increasing this number decreases the amount of beatmap sections in general. Note that this value does not account for the speed multiplier of
     * the play, similar to the way replay object data is stored.
     * @param maxSectionDeltaTime The minimum object count required to make a beatmap section. Increasing this number decreases the amount of beatmap sections.
     */
    static generateSections(
        calculator: DifficultyCalculator | RebalanceDifficultyCalculator,
        minSectionObjectCount: number,
        maxSectionDeltaTime: number
    ): BeatmapSection[] {
        const beatmapSections: BeatmapSection[] = [];
        let firstObjectIndex: number = 0;

        for (let i = 0; i < calculator.objects.length - 1; ++i) {
            const current: DifficultyHitObject | RebalanceDifficultyHitObject =
                calculator.objects[i];
            const next: DifficultyHitObject | RebalanceDifficultyHitObject =
                calculator.objects[i + 1];

            const realDeltaTime: number =
                next.object.startTime - current.object.endTime;

            if (realDeltaTime >= maxSectionDeltaTime) {
                // Ignore sections that don't meet object count requirement.
                if (i - firstObjectIndex < minSectionObjectCount) {
                    firstObjectIndex = i + 1;
                    continue;
                }

                beatmapSections.push({
                    firstObjectIndex,
                    lastObjectIndex: i,
                });

                firstObjectIndex = i + 1;
            }
        }

        // Don't forget to manually add the last beatmap section, which would otherwise be ignored.
        if (
            calculator.objects.length - firstObjectIndex >
            minSectionObjectCount
        ) {
            beatmapSections.push({
                firstObjectIndex,
                lastObjectIndex: calculator.objects.length - 1,
            });
        }

        return beatmapSections;
    }
}
